<?php

namespace lib;

use lib\help\Arr;
use lib\trail\Instance;
use lib\help\Env;
use PDO;

/**
 * Class Model
 * @package lib
 */
class Model
{
    use Instance;

    /**
     * @var PDO
     */
    protected $pdo;

    /**
     * @var
     */
    private $prepare;

    /**
     * 数据库配置
     * @var array
     */
    protected $database = [
        'type' => 'MYSQL',
        'hostname' => '127.0.0.1',
        'database' => '',
        'username' => '',
        'password' => '',
        'hostport' => '',
        'prefix' => 'eb_',
    ];

    /**
     * Model constructor.
     * @param array $config
     */
    private function __construct()
    {
        $this->database['type'] = Env::get('database.type');
        $this->database['hostname'] = Env::get('database.hostname');
        $this->database['database'] = Env::get('database.database');
        $this->database['username'] = Env::get('database.username');
        $this->database['password'] = Env::get('database.password');
        $this->database['hostport'] = Env::get('database.hostport');
        $this->database['prefix'] = Env::get('database.prefix', 'eb_');
        $this->connect();
    }

    /**
     * 设置表前缀
     * @param string $prefix
     * @return static
     */
    public static function setTablePrefix(string $prefix)
    {
        self::instance()->database['prefix'] = $prefix;
        return self::instance();
    }

    /**
     * 开启事务
     * @return bool
     */
    public static function beginTrans()
    {
        return self::instance()->pdo->beginTransaction();
    }

    /**
     * 提交事务
     * @return bool
     */
    public static function commit()
    {
        return self::instance()->pdo->commit();
    }

    /**
     * 回滚事务
     * @return bool
     */
    public static function rollBack()
    {
        return self::instance()->pdo->rollBack();
    }

    public function __clone()
    {
        trigger_error('Clone is not allow!', E_USER_ERROR);
    }

    /**
     * 链接mysql
     */
    private function connect()
    {
        try {
            $dataBaseAddress = $this->database['type'] . ':host=' . $this->database['hostname'] . ';dbname=' . $this->database['database'] . ';charset=utf8;port=' . $this->database['hostport'];
            $this->pdo = new PDO($dataBaseAddress, $this->database['username'], $this->database['password']);
            $this->pdo->setAttribute(PDO::ATTR_PERSISTENT, true);
            //设置抛出错误
            $this->pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
            //设置当字符串为空时转换为sql的null
            $this->pdo->setAttribute(PDO::ATTR_ORACLE_NULLS, true);
            //由MySQL完成变量的转义处理
            $this->pdo->setAttribute(PDO::ATTR_EMULATE_PREPARES, false);
        } catch (\Throwable $e) {
            $this->msg($e);
        }
    }

    /**
     * @param string $theError
     */
    private function msg($theError = null)
    {
        if ($theError instanceof \Throwable) {
            $strTheError = [
                '错误信息' => $theError->getMessage() . "<br/>",
                '错误行数' => $theError->getLine() . '<br/>',
                '错误码' => $theError->getCode() . '<br/>',
                '错误文件' => $theError->getFile() . '<br/>'
            ];
        } else {
            $strTheError = $theError;
        }
        $strTheError = print_r($strTheError, true);
        $html = "<html>
                  <head>
                    <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>
                    <title>Mysql error</title>
                  </head>
                  <body>
                    <div style='width: 50%; height: 200px; border: 1px solid red; font-size: 12px;'>
                       $strTheError
                    </div>
                  </body>
               </html>
               ";
        echo $html;
        exit;
    }

    /**
     * @param $table
     * @param $row
     * @return mixed
     */
    public function insert(string $table, $row, bool $isSql = false)
    {
        $sql = $this->sqlCreate("insert", $table, $row);
        if ($isSql) {
            return $sql;
        }
        $result = $this->sqlExec($sql);
        return $result;
    }

    /**
     * @param $table
     * @param $row
     * @param $where
     */
    public function update(string $table, $row, $where)
    {
        $sql = $this->sqlCreate("update", $table, $row, $where);
        $result = $this->sqlExec($sql);
    }

    /**
     * @param $table
     * @param $where
     */
    public function delete(string $table, $where)
    {
        $sql = $this->sqlCreate("delete", $table, "", $where);
        $result = $this->sqlExec($sql);
    }

    /**
     * @param $action
     * @param $table
     * @param string $row
     * @param string $where
     * @return string
     */
    private function sqlCreate(string $action, string $table, $row = "", $where = "")
    {
        $actionArr = [
            "insert" => "INSERT INTO ",
            "update" => "UPDATE ",
            "delete" => "DELETE FROM "
        ];
        if ($action == 'insert') {
            if (is_array($row)) {
                $fields = array_keys($row);
                $value = array_values($row);
                $values = '';
                foreach ($value as $item) {
                    $values .= "\"" . $item . "\",";
                }
                $row = '(`' . implode('`,`', $fields) . '`) VALUES (' . rtrim($values, ',') . ')';
            }
        } else {
            $row = empty($row) ? "" : $this->rowCreate($row);
        }
        $where = empty($where) ? "" : " WHERE " . $where;
        $sql = $actionArr[$action] . '`' . $table . '`' . $row . $where;
        return $sql;
    }

    /**
     * @param $row
     * @return string
     */
    private function rowCreate($row)
    {
        $sqlrow = " SET";
        foreach ($row as $key => $val) {
            $sqlrow .= " " . $key . "='" . $val . "',";
        }
        return trim($sqlrow, ",");
    }


    /**
     * @param $sql
     * @return mixed
     */
    private function sqlExec($sql)
    {
        try {
            $result = $this->pdo->exec(str_replace('\\', '\\\\', $sql));
        } catch (PDOException $e) {
            $this->msg($e->getMessage());
        }
        return $result;
    }

    /**
     * @return mixed
     */
    public function lastinsertid()
    {
        return $this->pdo->lastinsertid();
    }

    /**
     * @param $table
     * @param string $fields
     * @param string $where
     * @param string $orderby
     * @param string $sort
     * @param string $limit
     * @return $this
     */
    public function selectSql($table, $fields = "", $where = "", $orderby = "", $sort = "", $limit = "")
    {
        $fields = empty($fields) ? "*" : $fields;
        $sqlSelect = $this->sqlCreateSelect($table, $fields, $where, $orderby, $sort, $limit);
        $this->query($sqlSelect, $where);
        return $this;
    }

    /**
     * @param $table
     * @param string $fields
     * @param string $where
     * @return $this
     */
    public function updateSql($table, $data, $where = "")
    {
        $whereSql = empty($where) ? " 1=1 " : $this->whereCreate($where);
        $row = '';
        if (is_array($data)) {
            $row = $this->rowCreate($data);
        } else {
            $row = ' set ' . $data;
        }
        $sql = "UPDATE " . '`' . $table . '`' . $row . " WHERE " . $whereSql;
        return $this->query($sql, $where);
        return $this;
    }

    /**
     * @param $table
     * @param string $fields
     * @param string $where
     * @param string $orderby
     * @param string $sort
     * @param string $limit
     * @return string
     */
    private function sqlCreateSelect($table, $fields = "*", $where = "", $orderby = "", $sort = "", $limit = "")
    {
        $whereSql = empty($where) ? " 1=1 " : $this->whereCreate($where);
        $orderbySql = empty($orderby) ? "" : " ORDER BY " . $orderby . " " . $sort;
        $limitSql = empty($limit) ? "" : " LIMIT " . $limit;
        $sql = "SELECT $fields FROM $table WHERE " . $whereSql . $orderbySql . $limitSql;
        return $sql;
    }

    /**
     * @param $where
     * @return string
     */
    private function whereCreate($where)
    {
        $whereSql = "";
        foreach ($where as $key => $val) {
            if (is_array($val)) {
                foreach ($val as $k => $v) {
                    $whereSql .= " `" . $key . '` ' . $k . ' :' . $key . " AND";
                }
            } else {
                $whereSql .= " " . $key . "=:" . $key . " AND";
            }
        }
        return $whereSql . " 1=1 ";
    }

    /**
     * @param $sql
     * @param $where
     */
    private function query($sql, $where)
    {
        try {
            $this->prepare = $this->pdo->prepare($sql);

        } catch (PDOException $e) {
            $this->msg("预处理sql出错信息:" . $e->getMessage() . "<br>sql:(" . $sql . ")");
        }
        empty($where) ? "" : $this->bind($where);

        $this->prepare->execute();
    }

    /**
     * @param $where
     */
    private function bind($where)
    {
        foreach ($where as $key => $val) {
            $this->prepare->bindValue(":" . $key, is_array($val) ? array_values($val)[0] : $val);
        }
    }

    /**
     * @return mixed
     */
    public function selectFind()
    {
        $result = $this->prepare->fetch(PDO::FETCH_ASSOC);
        return $result;
    }

    /**
     * @return mixed
     */
    public function selectAll()
    {
        $result = $this->prepare->fetchAll(PDO::FETCH_ASSOC);
        return $result;
    }

    /**
     * @return mixed
     */
    public function selectCount()
    {
        $total = $this->prepare->rowCount();
        return $total;
    }

}
